from kathairo.tokenization.MaximalMatchingTokenizer import MaximalMatchingTokenizer
import os.path
import os
from typing import Optional
from importlib.resources import files

# <summary>
# A tokenizer for bible translations (not general Chinese translations, see below).
# 
# Chinese words and combination corrections provided by Andi Wu (andi.wu@globalbibleinitiative.org)
# </summary>

class ChineseBibleWordTokenizer(MaximalMatchingTokenizer):
   

    def __init__(self,chineseTokenizerDataDirectoryPath: Optional[str] = None,max_gram: Optional[int] = None):
        if max_gram is None:
            max_gram = 10
        self.WORDS_FILE_NAME = "words.txt"
        self.COMBINATION_CORRECTIONS_FILE_NAME = "combination_corrections.txt"
        super().__init__(max_gram)
            
        words_path = files('kathairo.tokenization.Data.ChineseBibleWordTokenizer') / self.WORDS_FILE_NAME
        with words_path.open('r', encoding='utf-8') as words_file:
            #lines = words_file.readlines()
            #for line in lines:
            #    if line is not "":
            #        super().Words.Add
                    
            self.Words.update(line.strip() for line in words_file if line.strip())

        corrections_path = files('kathairo.tokenization.Data.ChineseBibleWordTokenizer') / self.COMBINATION_CORRECTIONS_FILE_NAME
        with corrections_path.open('r', encoding='utf-8') as corrections_file:
            #lines = corrections_file.readlines()
            #for line in lines:
            #    parts = line.split("\t")
            #    if parts[0] not in super().CombinationCorrection:
            #        super().CombinationCorrection.Add(parts[0], parts[1])
            
            for line in corrections_file:
                parts = line.strip().split("\t")
                if parts[0] not in self.CombinationCorrections:
                    self.CombinationCorrections[parts[0]] = parts[1]


from pathlib import Path
from typing import Optional

from machine.scripture.verse_ref import Versification
from machine.utils.typeshed import StrPath
from machine.corpora.file_stream_container import FileStreamContainer
from machine.corpora.stream_container import StreamContainer
from machine.corpora.usfm_stylesheet import UsfmStylesheet
from kathairo.parsing.usfm.usfm_text_base import ModifiedUsfmTextBase
from machine.corpora.usfm_parser_handler import UsfmParserHandler


class UsfmFileText(ModifiedUsfmTextBase):
    def __init__(
        self,
        stylesheet: UsfmStylesheet,
        encoding: str,
        filename: StrPath,
        handler: UsfmParserHandler,
        psalmSuperscriptionTag: str, 
        versification: Optional[Versification] = None,
        include_markers: bool = False,
    ) -> None:
        super().__init__(_get_id(filename, encoding), stylesheet, encoding, handler, psalmSuperscriptionTag, versification, include_markers) #passes in handler

        self._filename = Path(filename)

    def _create_stream_container(self) -> StreamContainer:
        return FileStreamContainer(self._filename)


def _get_id(filename: StrPath, encoding: str) -> str:
    with open(filename, "r", encoding=encoding) as file:
        for line in file:
            line = line.strip()
            if line.startswith("\\id "):
                id = line[4:]
                index = id.find(" ")
                if index != -1:
                    id = id[:index]
                return id.strip().upper()
    raise RuntimeError("The USFM does not contain an 'id' marker.")



from machine.tokenization import WhitespaceTokenizer
from machine.annotations.range import Range
from typing import List

'''
Copyright © 2023 Cherith Analytics, LLC
Permission is hereby granted, free of charge, to any person obtaining a copy of portions of this file, specifically WhitespaceDelimitLongestWordMatches() and GetNgram(), (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

''' Notes from Andi:
The maximal matching algorithm has two problems: 

(1) it can be wrong when there are combinational ambiguity 
where a sequence of characters can be single word/token in some contexts but separate words in other contexts.  
Since MM always go for the bigger one, these sequences will always be segmented as a single word regardless of contexts.  
Cases of combinational ambiguity do not occur very frequently, though.

(2) it can be wrong when there are overlapping ambiguities where a sequence of characters such as ABC can be 
segmented as AB C or A BC depending on the contexts.  The overlap.txt file is used to correct such mistakes, 
favoring the more likely segmentation in the Bible.

Another way to resolve overlapping ambiguities is to do it from both directions.  The code I gave you only does 
it in one direction: left to right.  But you can do both left to right and right to left.  The tokens you can get in 
both left and right directions are guaranteed to be correct, but you need a separate routine to handle the ones that 
do not agree in both directions.  You may lose words if you only keep the tokens that you can get from both directions.  
In other words, bi-directional maximal matching can improve precision but lose in recall.

In spite of these problems, the accuracy of MM is pretty high and meets the needs of most applications.

To improve the coverage of the current tokenizer (i.e. to be able to handle texts that contain words that are not in the word list), 
we just need to add words to the list.  However, the more words are in the list, the more likely we will have combinational and overlapping 
ambiguities.
'''

# <summary>
#
# Based on Andi Wu's BibleTokenizer (andi.wu@globalbibleinitiative.org)
# 
# 1. looks at a verse's characters as if each is a linear (sequential) graph of 
# vertices (characters) and edges (their pairwise adjacency).
# 2. tries to match them to the pairs of characters in the vocabulary file (words.txt), considering each pair 
# in this file as as two vertices connected by a pair-wise-edge, using a simple greedy algorithm,
# 3. to find the maximal matching set for the verse consistent with the pairs in the vocabulary file,
# 4. which are considered characters joined together as a token.
# 
# Words contains all the words to match.
# 
# CombinationCorrections map combination errors to corrections. 
# 
# </summary>

class MaximalMatchingTokenizer(WhitespaceTokenizer):
      
    MAX_GRAM_DEFAULT:int = 10

    _maxGram:int

    #CombinationCorrections = {}

    # <summary>
    # Add words to maximal match on to Words property.
    # </summary>
    # <param name="maxGram"></param>
    def __init__(self, maxGram:int):
            super().__init__()
            self._maxGram = maxGram
            self.CombinationCorrections = {}
            self.Words = set()

    def tokenize_as_ranges(self, data:str, range:Range)-> List[Range]:
        if range is None:
            range = Range.create(0, len(data))
        
        #segment words
        whitespaceSegmentedText = self.WhitespaceDelimitLongestWordMatches(data[range.start:range.length], self.Words, self._maxGram)

        #correct overlaps
        self.CorrectCombinations(whitespaceSegmentedText)

        rangesForWhitespaceSegmentedText = super().tokenize_as_ranges(whitespaceSegmentedText, Range.create(0, len(whitespaceSegmentedText)))

        # Adjust the ranges for the original text
        adjusted_ranges = [Range.create(r.start - i, r.end - i) for i, r in enumerate(rangesForWhitespaceSegmentedText)]

        return adjusted_ranges

    # <summary>
    # Use to correct the following combination errors by replacing a sequence of characters with another:
    # 1. maximal match algorithm favors largest combination of characters, but this can be incorrect when in context with other characters
    # that follow. For example, if algorithm tokenizes as AB, in cases where AB is followed by C the characters A and B should instead be tokenized as A B.
    # 2. When there is an overlapping ambiguity, for example, if algorithm tokenizes as AB C but it could have also tokenized as A BC.
    # 
    # These corrections should favor the majority cases.
    # </summary>
    # <param name="whitespaceSegmentedText"></param>
    
    def CorrectCombinations(self, whitespaceSegmentedText):
        for key, value in self.CombinationCorrections.items():
            whitespaceSegmentedText = whitespaceSegmentedText.replace(key, value)

    # <summary>
    # Note: whitespace delimits spaces that aren't a part of a match.
    # </summary>
    # <param name="text"></param>
    # <param name="words"></param>
    # <param name="maxGram"></param>
    # <returns></returns>
    def WhitespaceDelimitLongestWordMatches(self, text:str, words:[str], maxGram:int):
        segments = ""

        chars = list(text)

        #for i in range(0,i<chars.length,1):
        #    for n in range(maxGram,n > 0,-1):
        #        ngram = self.GetNgram(chars, i, n)
        #        if ngram != "" and (n == 1 or ngram in self.Words):
        #            segments += ngram + (" " if len(ngram) != 1 or not ngram.isspace() else "") # only delimit if the single character is not a space, otherwise just add space.
        #            i = i + n - 1; # need to subtract 1 because the iterator increments after the body.
        #            break
        
        i=0
        while i < len(chars):
            for n in range(self._maxGram, 0, -1):
                ngram = self.GetNgram(chars, i, n)
                if ngram != "" and (n == 1 or ngram in self.Words):
                    segments += ngram + (" " if len(ngram) != 1 or not ngram.isspace() else "")
                    i = i + n - 1
                    break
            i += 1
        
        return segments

    def GetNgram(self, chars:[str], index:int, length:int):
        ngram:str = ""

        i:int = index
        while(i+length <= len(chars) and length>0):
                
                ngram += chars[i]
                
                i+=1
                length-=1

        #for i in range(index, index + length):
        #    if i < len(chars):
        #        ngram += chars[i]

        return ngram

    #def get_ngram(self, chars, index, length):
    #    if index + length <= len(chars) and length > 0:
    #        return "".join(chars[index:index+length])
    #    return ""


from __future__ import annotations

import xml.etree.ElementTree as etree
from typing import Iterable

from machine.scripture.verse_ref import are_overlapping_verse_ranges
from machine.corpora.corpora_utils import merge_verse_ranges
from machine.corpora.usx_verse import UsxVerse
from machine.corpora.usx_verse_parser import UsxVerseParser
from machine.corpora.usx_verse_parser import _ParseContext
from machine.corpora.usx_verse_parser import _is_numbered_style


class ModifiedUsxVerseParser(UsxVerseParser):
    def __init__(self, merge_segments: bool = False) -> None:
        self._merge_segments = merge_segments
    
    def _parse_element(self, elem: etree.Element, ctxt: _ParseContext) -> Iterable[UsxVerse]:
        if elem.text is not None and ctxt.chapter is not None and ctxt.verse is not None:
            ctxt.add_token(elem.text)
        for e in elem:
            if e.tag == "chapter":
                if ctxt.chapter is not None and ctxt.verse is not None:
                    yield ctxt.create_verse()
                ctxt.chapter = e.get("number")
                ctxt.verse = None
                ctxt.is_sentence_start = True
            elif e.tag == "para":
                if not _is_verse_para(e):
                    ctxt.is_sentence_start = True
                    continue
                
                if(e.get("style", "") == "d" and (ctxt.verse != None and ctxt.verse != "0")):
                    continue
                
                # include superscriptions in text (TODO, limit to just the Psalms)
                if (e.get("style", "") == "d" and ctxt.chapter != '119'):#or e.get("style", "") == "s"
                    verse = "0"
                    ctxt.verse = verse
                ###
                        
                ctxt.para_element = e
                for evt in self._parse_element(e, ctxt):
                    yield evt
            elif e.tag == "verse":
                if "eid" in e.attrib:
                    yield ctxt.create_verse()
                    ctxt.verse = None
                else:
                    verse = e.get("number")
                    if verse is None:
                        verse = e.get("pubnumber")
                    assert verse is not None
                    if ctxt.chapter is not None and ctxt.verse is not None:
                        if verse == ctxt.verse:
                            yield ctxt.create_verse()

                            # ignore duplicate verse
                            ctxt.verse = None
                        elif are_overlapping_verse_ranges(verse, ctxt.verse):
                            # merge overlapping verse ranges in to one range
                            ctxt.verse = merge_verse_ranges(verse, ctxt.verse)
                        else:
                            yield ctxt.create_verse()
                            ctxt.verse = verse
                    else:
                        ctxt.verse = verse
            elif e.tag == "char":
                if e.get("style") == "rq":
                    if ctxt.chapter is not None and ctxt.verse is not None:
                        ctxt.add_token("", e)
                else:
                    for evt in self._parse_element(e, ctxt):
                        yield evt
            elif e.tag == "wg":
                if e.text is not None and ctxt.chapter is not None and ctxt.verse is not None:
                    ctxt.add_token(e.text, e)
            elif e.tag == "figure":
                if ctxt.chapter is not None and ctxt.verse is not None:
                    ctxt.add_token("", e)

            if e.tail is not None and ctxt.chapter is not None and ctxt.verse is not None:
                ctxt.add_token(e.tail)


_NONVERSE_PARA_STYLES = {"ms", "mr", "sr", "r", "sp", "rem", "restore", "cl", "s"} # removed "d" and "s" to include superscriptions

def _is_verse_para(para_elem: etree.Element) -> bool:
    style = para_elem.get("style", "")
    if style in _NONVERSE_PARA_STYLES:
        return False

    if _is_numbered_style("ms", style):
        return False

    if _is_numbered_style("s", style):
        return False

    return True



from pathlib import Path
from typing import List, Optional

from machine.scripture.verse_ref import Versification
from machine.utils.typeshed import StrPath
from machine.corpora.corpora_utils import get_usx_versification
from machine.corpora.scripture_text_corpus import ScriptureTextCorpus
from kathairo.parsing.usx.usx_file_text import UsxFileText #use Modified UsxFileText


class UsxFileTextCorpus(ScriptureTextCorpus):
    def __init__(
        self,
        project_dir: StrPath,
        versification: Optional[Versification] = None,
    ) -> None:
        project_dir = Path(project_dir)
        versification = get_usx_versification(project_dir, versification)
        texts: List[UsxFileText] = []
        for filename in project_dir.glob("*.usx"):
            texts.append(UsxFileText(filename, versification))
        super().__init__(versification, texts)



def get_file_location(directory:str, language:str, project_name:str, sub_directory:str, file_prefix:str):
    output_file_location = directory+"/"+language+"/"+project_name+"/"+sub_directory+"/"+file_prefix+"_"+project_name+".tsv"
    return output_file_location

import importlib.util
import sys

def import_module_from_path(module_name, file_path):
    if(file_path is None):
        return None
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


from pathlib import Path
from typing import Optional

from machine.scripture.verse_ref import Versification
from machine.utils.typeshed import StrPath
from machine.corpora.corpora_utils import get_usx_id
from machine.corpora.file_stream_container import FileStreamContainer
from machine.corpora.stream_container import StreamContainer
from kathairo.parsing.usx.usx_text_base import UsxTextBase #use modified UsxTextBase


class UsxFileText(UsxTextBase):
    def __init__(self, filename: StrPath, versification: Optional[Versification] = None) -> None:
        self._filename = Path(filename)
        super().__init__(get_usx_id(self._filename), versification)

    def _create_stream_container(self) -> StreamContainer:
        return FileStreamContainer(self._filename)



import regex as re

class DefaultRegexRules:
    INNER_WORD_PUNCT = r"[&\-:=@\xAD\xB7\u2010\u2011\u2027]+|['_]+"
    INNER_WORD_PUNCT_REGEX = re.compile(rf"{INNER_WORD_PUNCT}")

    NUMBER_COMMA_REGEX = re.compile(
        r"(?<=\d),(?=\d)"
    )

    NUMBER_PERIOD_REGEX = re.compile(
        r"(?<=\d)\.(?=\d)"
    )

    RIGHT_SINGLE_QUOTE_AS_APOSTROPHE_REGEX = re.compile(
        r"(?<=\p{L})’(?=\p{L})"
    )

    CONTRACTION_WORD_REGEX = re.compile(
        r"\b\w+(?:[\'\w\’]+)?\b"
    )
    
    def get_regex_rules(self):
        regex_rules = [
            self.INNER_WORD_PUNCT_REGEX,
            self.NUMBER_COMMA_REGEX,
            self.NUMBER_PERIOD_REGEX,
            self.RIGHT_SINGLE_QUOTE_AS_APOSTROPHE_REGEX,
            #CONTRACTION_WORD_REGEX
        ]
        
        return regex_rules

    #NON_JOINING_PUNCT =r"[.،«?!।၊–…{}—《》（）‘’“”;？：；。！，、,\[\]()]"
    NON_JOINING_PUNCT = r"[.،«?!।၊–…{}—《》（）‘’“”;？：；。！，、,\[\]()]"
    #does nothing, oddly        ၊ 
    #is joining in some cases   –   —

    WORD_LEVEL_PUNCT_REGEX = re.compile(
        fr"(?<=\w)(\p{{P}}(?<!{NON_JOINING_PUNCT}))(?=\w)"
    )


import polars as pl

df = pl.read_csv("nt_CUVMPS_minimum.tsv", separator='\t', infer_schema_length=0)

df = df.with_columns(pl.col("id").str.slice(0, 8))

result = df.group_by("id").agg([
    pl.col("text").str.concat("")
])

result = result.sort("id")

output_file_name = "nt_CUVMPS_verse.tsv"

result.write_csv(output_file_name, separator='\t')


from typing import Generator, Optional

from machine.scripture.verse_ref import Versification
from machine.corpora.corpora_utils import gen
from machine.corpora.text_row import TextRow
from machine.corpora.usfm_parser import parse_usfm
from machine.corpora.usfm_parser_handler import UsfmParserHandler
from machine.corpora.usfm_stylesheet import UsfmStylesheet
from machine.corpora.usfm_parser_handler import UsfmParserHandler
from machine.corpora.usfm_text_base import _TextRowCollector
from machine.corpora.usfm_text_base import UsfmTextBase

class ModifiedUsfmTextBase(UsfmTextBase):
    def __init__(
        self,
        id: str,
        stylesheet: UsfmStylesheet,
        encoding: str,
        handler: UsfmParserHandler,
        psalmSuperscriptionTag: str,
        versification: Optional[Versification],
        include_markers: bool,
    ) -> None:
        super().__init__(id, stylesheet, encoding, versification, include_markers)

        self._stylesheet = stylesheet
        self._encoding = encoding
        self.handler = handler #passes in handler
        self.psalm_superscription_tag = psalmSuperscriptionTag
        self._include_markers = include_markers

    def _get_rows(self) -> Generator[TextRow, None, None]:
        usfm = self._read_usfm()
        row_collector = _TextRowCollector(self)
        if(self.handler is not None): #uses handler if not None
            row_collector = self.handler(self, psalm_superscription_tag = self.psalm_superscription_tag)
        parse_usfm(usfm, row_collector, self._stylesheet, self.versification, preserve_whitespace=self._include_markers)
        return gen(row_collector.rows)



class Meta:

    def __init__(self, process):
        self.process = process

class Alignment:

    def __init__(self, id, source_ids, target_ids, meta):
        self.id = id
        self.source_ids = source_ids
        self.target_ids = target_ids
        self.meta = meta

    def addAlignedPair(self, source_id, target_id):
        self.source_ids.append(source_id)
        self.target_ids.append(target_id)

class AlignmentSet:

    def __init__(self, alignments):
        self.alignments = alignments

    def addAlignment(self, alignment):
        self.alignments.append(alignment)

    def removeAlignment(self, alignment):
        self.alignments.remove(alignment)

    def fetchAlignmentById(self, id):

        fetchedAlignment = next ( 
            (alignment for alignment in self.alignments if alignment.id == id), 
            None)
        
        return fetchedAlignment
    
    def fetchAlignmentBySourceId(self, source_id):

        fetchedAlignment = next ( 
            (alignment for alignment in self.alignments if alignment.source_ids.contains(source_id)), 
            None)
        
        return fetchedAlignment
    
    def fetchAlignmentByTargetId(self, target_id):

        fetchedAlignment = next ( 
            (alignment for alignment in self.alignments if alignment.target_ids.contains(target_id)), 
            None)
        
        return fetchedAlignment

class AlignmentSetManager:

    def __init__(self, alignmentSet):
        self.alignmentSet = alignmentSet

    

def convert_to_dict(obj):
    #if isinstance(obj, (Meta, Alignment)):
    return obj.__dict__
    #return obj


import json

meta = Meta("manual")

alignment = Alignment("66022021.9", ["66022021007"], ["660220210131"], meta)

json_str = json.dumps(alignment, default=convert_to_dict, indent=2)

print(json_str)
   


import csv
from pathlib import Path
import os

def reconstitute(tsv_file_path:Path, project_name:str, language:str):
    #directory = Path(sys.argv[1])
    #input_files = directory.glob("*reconstitution*.tsv")
    input_files = [Path(tsv_file_path)]
    for input_file in input_files:
        
        verses = create_verse_array(input_file)

        output_file = input_file.parent.parent.parent / project_name / "reconstituted" / f"{input_file.stem}_reconstitution.tsv"
        
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        with open(output_file, 'w', newline='', encoding='utf-8') as out_file:
            tsv_writer = csv.writer(out_file, delimiter='\t', quoting=csv.QUOTE_NONE, quotechar=None)

            tsv_writer.writerows(verses) #OLD WAY

def create_verse_array(input_file):
        rows = [r for r in csv.DictReader(input_file.open("r", encoding='utf-8'), delimiter="\t", quoting=csv.QUOTE_NONE, quotechar=None)]
        verses = []
        verses.append(["id", "text"])
        text = ""
        parts = []
        last_verse = None
        for row in rows:
            this_verse = row["id"][0:8]
            if last_verse is None:
                last_verse = this_verse
                parts.append(f"{this_verse}")

            if last_verse != this_verse:
                parts.append(text)
                verses.append(parts)#.strip()
                
                last_verse = this_verse
                
                text = ""
                parts = [f"{this_verse}"]#[f""]#

            text+=(row["text"])
            if row["skip_space_after"] == "y":
                continue
            text+=(" ")
        parts.append(text)
        verses.append(parts)#.strip()
        
        return verses


from typing import Iterable, Optional

from machine.annotations.range import Range
from machine.tokenization import StringTokenizer

from machine.tokenization import WhitespaceTokenizer

class WhitespaceIncludedTokenizer(StringTokenizer):
    def tokenize_as_ranges(self, data: str, data_range: Optional[Range[int]] = None) -> Iterable[Range[int]]:
        if data_range is None:
            data_range = Range.create(0, len(data))
        start_index = -1
        for i in data_range:
            if self._is_whitespace(data[i]):
                if start_index != -1:
                    yield Range.create(start_index, i)
                yield Range.create(i, i+1) # Added for kathairo.py to include whitespace
                start_index = -1
            elif start_index == -1:
                start_index = i

        if start_index != -1:
            yield Range.create(start_index, data_range.end)

    def _is_whitespace(self, c: str) -> bool:
        return c.isspace() or c == "\u200b" or c == "\ufeff"


WHITESPACE_TOKENIZER = WhitespaceTokenizer()



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Optional, Tuple, cast

import regex as re

from machine.annotations.range import Range
from machine.utils.string_utils import is_control, is_punctuation, is_symbol
from kathairo.tokenization.whitespace_included_tokenizer import WhitespaceIncludedTokenizer
from spacy.lang.fr.tokenizer_exceptions import FR_BASE_EXCEPTIONS

URL_REGEX = re.compile(r"(?:[\w-]+://?|www[.])[^\s()<>]+(?:[\w\d]+|(?:[^\p{P}\s]|/))", re.IGNORECASE)

CONTRACTION_WORD_REGEX = re.compile(r"\b\w+([-]\w+)*[\'\’]\w+\b")

class LatinWhitespaceIncludedWordTokenizer(WhitespaceIncludedTokenizer): #uses WhitespaceIncludedTokenizer
    def __init__(self, regex_rules_class, abbreviations: Iterable[str] = [], treat_apostrophe_as_single_quote: bool = False, language:str = None) -> None:
        self._abbreviations = {a.lower() for a in abbreviations}
        self.treat_apostrophe_as_single_quote = treat_apostrophe_as_single_quote
        self.language = language
        self.regex_rules = regex_rules_class.get_regex_rules()

    def tokenize_as_ranges(self, data: str, data_range: Optional[Range[int]] = None) -> Iterable[Range[int]]:
        if data_range is None:
            data_range = Range.create(0, len(data))
        ctxt = LatinWhitespaceIncludedWordTokenizer._TokenizeContext()
        for char_range in super().tokenize_as_ranges(data, data_range):
            url_match = URL_REGEX.match(data[char_range.start : char_range.end])
            if url_match is not None:
                url_len = len(url_match.group())
                yield Range.create(char_range.start, char_range.start + url_len)
                ctxt.index = char_range.start + url_len
            else:
                ctxt.index = char_range.start

            ctxt.word_start = -1
            ctxt.inner_word_punct = -1

            while ctxt.index < char_range.end:
                token_range1, token_range2 = self._process_character(data, data_range, ctxt)
                if token_range1 is not None:
                    yield token_range1
                if token_range2 is not None:
                    yield token_range2

            if ctxt.word_start != -1:
                if ctxt.inner_word_punct != -1:
                    inner_punct_str = data[ctxt.inner_word_punct : char_range.end]
                    if (
                        inner_punct_str == "." and self._is_abbreviation(data, ctxt.word_start, ctxt.inner_word_punct)
                    ) or (inner_punct_str == "'" and not self.treat_apostrophe_as_single_quote):
                        #range1 = data[ctxt.word_start, char_range.end]
                        yield Range.create(ctxt.word_start, char_range.end)
                    else:
                        #range1 = data[cast(int, ctxt.word_start): ctxt.inner_word_punct]
                        yield Range.create(cast(int, ctxt.word_start), ctxt.inner_word_punct)
                        #range2 = data[ctxt.inner_word_punct: char_range.end]
                        yield Range.create(ctxt.inner_word_punct, char_range.end)
                else:
                    #range1 = data[ctxt.word_start: char_range.end]
                    yield Range.create(ctxt.word_start, char_range.end)

    def _process_character(
        self, data: str, data_range: Range[int], ctxt: LatinWhitespaceIncludedWordTokenizer._TokenizeContext
    ) -> Tuple[Optional[Range[int]], Optional[Range[int]]]:
        token_ranges: Tuple[Optional[Range[int]], Optional[Range[int]]] = (None, None)
        c = data[ctxt.index]
        end_index = ctxt.index + 1

        if is_punctuation(c) or is_symbol(c) or is_control(c):
            
            while end_index != data_range.end and data[end_index] == c:
                end_index += 1
            if ctxt.word_start == -1:
                if c == "'" and not self.treat_apostrophe_as_single_quote:
                    ctxt.word_start = ctxt.index
                else:
                    match = None
                    for rule in self.regex_rules:
                        match = rule.match(data, ctxt.index)
                    if match is None:
                        #range1 = data[ctxt.index: end_index]
                        token_ranges = (Range.create(ctxt.index, end_index), None)
                    else:
                        ctxt.word_start = ctxt.index
            elif ctxt.inner_word_punct != -1:
                inner_punct_str = data[ctxt.inner_word_punct : ctxt.index]
                if inner_punct_str == "'" and not self.treat_apostrophe_as_single_quote:
                    #range1 = data[ctxt.word_start: ctxt.index]
                    token_ranges = (Range.create(ctxt.word_start, ctxt.index), None)
                else:
                    #range1 = data[ctxt.word_start: ctxt.inner_word_punct]
                    #range2 = data[ctxt.inner_word_punct: ctxt.index]
                    token_ranges = (
                        Range.create(ctxt.word_start, ctxt.inner_word_punct),
                        Range.create(ctxt.inner_word_punct, ctxt.index),
                    )
                ctxt.word_start = ctxt.index
            else:
                for rule in self.regex_rules:
                    substring = data[ctxt.index-1:ctxt.index+2]      
                    
                    match = rule.match(data, ctxt.index)
                    #match = rule.search(substring)
                    if match is not None:
                        ctxt.inner_word_punct = ctxt.index
                        group = match.group()
                        ctxt.index += len(group)
                    
                        if(self.language == "fra"):
                            contraction_token = CONTRACTION_WORD_REGEX.match(data, ctxt.word_start)
                            if(contraction_token is not None):
                                group = contraction_token.group().replace("’","'")
                                if(group not in FR_BASE_EXCEPTIONS):
                                    #range1 = data[ctxt.word_start:ctxt.index]
                                    token_ranges = (Range.create(ctxt.word_start, ctxt.index),None)
                                    ctxt.word_start = -1
                                
                        return token_ranges
                #range1 = data[ctxt.word_start:ctxt.index]
                #range2 = data[ctxt.index:end_index]
                token_ranges = (Range.create(ctxt.word_start, ctxt.index), Range.create(ctxt.index, end_index))
                ctxt.word_start = -1
        elif ctxt.word_start == -1:
            ctxt.word_start = ctxt.index

        ctxt.inner_word_punct = -1
        ctxt.index = end_index
        return token_ranges

    def _is_abbreviation(self, data: str, start: int, end: int) -> bool:
        substr = data[start:end].lower()
        return substr in self._abbreviations

    @dataclass
    class _TokenizeContext:
        index: int = 0
        word_start: int = 0
        inner_word_punct: int = 0



from pathlib import Path
from typing import List, Optional

from machine.scripture.verse_ref import Versification, VersificationType
from machine.utils.typeshed import StrPath
from machine.corpora.scripture_text_corpus import ScriptureTextCorpus
from kathairo.parsing.usfm.usfm_file_text import UsfmFileText
from machine.corpora.usfm_stylesheet import UsfmStylesheet
from machine.corpora.usfm_parser_handler import UsfmParserHandler

class UsfmFileTextCorpus(ScriptureTextCorpus):
    def __init__(
        self,
        project_dir: StrPath,
        handler:UsfmParserHandler,
        psalmSuperscriptionTag:str,
        stylesheet_filename: StrPath = "usfm.sty",
        encoding: str = "utf-8-sig",
        versification: Optional[Versification] = None,
        include_markers: bool = False,
        file_pattern: str = "*.SFM",
    ) -> None:
        if versification is None:
            versification = Versification.get_builtin(VersificationType.ENGLISH)
        stylesheet = UsfmStylesheet(stylesheet_filename)
        texts: List[UsfmFileText] = []
        for sfm_filename in Path(project_dir).glob(file_pattern):
            texts.append(UsfmFileText(stylesheet, encoding, sfm_filename, handler, psalmSuperscriptionTag, versification, include_markers)) #passes in handler
        super().__init__(versification, texts)



import polars as pl

df = pl.read_csv("nt_CUVMPS.tsv", separator='\t', infer_schema_length=0)

results = df.select([pl.col("id"), pl.col("text")])

results.write_csv("nt_CUVMPS_minimum.tsv", separator='\t')


import unicodedata

def is_unicode_punctuation(char):
    category = unicodedata.category(char)
    return category.startswith("P")

def contains_number(string):
    return any(char.isdigit() for char in string)

empty_string = ""

zwsp = "​"#\u200b
zwj = "‍"#\u200d
zwnj = "‌"#\u200c
nbsp = "\xa0"
space = " "
six_per_em_space = " "#\u2006

stop_words = [
    space,
    zwsp,
    zwj,
    zwnj,
    nbsp,
    six_per_em_space
]

spaces = [
    space,
    nbsp,
    six_per_em_space
]

invisible_spaces = [
    zwsp
]




from typing import Iterable, List, Optional

from machine.scripture.verse_ref import VerseRef
from machine.corpora.text_row import TextRow
from machine.corpora.usfm_parser_state import UsfmParserState
from machine.corpora.usfm_token import UsfmToken, UsfmTokenType

from machine.corpora.usfm_text_base import UsfmTextBase
from machine.corpora.usfm_text_base import _TextRowCollector

class ModifiedTextRowCollector(_TextRowCollector):
    def __init__(self, text: UsfmTextBase, psalm_superscription_tag: str = "d") -> None:
        self._text = text
        self._rows: List[TextRow] = []
        self._verse_text = ""
        self._next_para_tokens: List[UsfmToken] = []
        self._verse_ref: Optional[VerseRef] = None
        self._sentence_start: bool = False
        self._next_para_text_started = False
        self._psalm_superscription_tag = psalm_superscription_tag

    @property
    def rows(self) -> Iterable[TextRow]:
        return self._rows

    def text(self, state: UsfmParserState, text: str) -> None:
        
        is_psalm_superscription = False
        
        if(state.prev_token is not None):
            is_psalm_superscription = ((state.prev_token.marker == self._psalm_superscription_tag) 
                                    and state.verse_ref.book == "PSA" 
                                    and (state.verse_ref.bbbcccvvvs != "019119000" and state.verse_ref.bbbcccvvvs != "019107000"))
        
        #includes superscription text
        if self is not None and is_psalm_superscription:
            self.verse(state, 0, "v", 0, 0)
    
        if self._verse_ref is None or (not state.is_verse_para and not is_psalm_superscription):
            return

        if self._text._include_markers:
            text = text.rstrip("\r\n")
            if len(text) > 0:
                if not text.isspace():
                    for token in self._next_para_tokens:
                        self._verse_text += str(token)
                    self._next_para_tokens.clear()
                    self._next_para_text_started = True
                self._verse_text += text
        elif (state.is_verse_text or is_psalm_superscription) and len(text) > 0:
            if (
                state.prev_token is not None
                and state.prev_token.type == UsfmTokenType.END
                and (self._verse_text == "" or self._verse_text[-1].isspace())
            ):
                text = text.lstrip()
            self._verse_text += text


import time
start = time.time()

import json
from concurrent.futures import ProcessPoolExecutor, as_completed

from kathairo.tokenization.ChineseBibleWordTokenizer import ChineseBibleWordTokenizer
from machine.tokenization import LatinWordTokenizer
from kathairo.tokenization.latin_whitespace_included_tokenizer import LatinWhitespaceIncludedWordTokenizer
from kathairo.tokenization.regex_rules import DefaultRegexRules

from kathairo.parsing.usx.usx_file_text_corpus import UsxFileTextCorpus
from kathairo.parsing.usfm.usfm_file_text_corpus import UsfmFileTextCorpus
from kathairo.parsing.usfm.usfm_handlers import ModifiedTextRowCollector

from machine.scripture import ORIGINAL_VERSIFICATION, Versification
from kathairo.helpers.paths import import_module_from_path

from kathairo.tsvs.build_tsv import corpus_to_tsv, tokens_to_tsv

import polars as pl

SOURCE_VERSIFICATION = Versification(name="sourceVersification", base_versification=ORIGINAL_VERSIFICATION)

def create_tokenizer(json_object, regex_rules_class):
    treat_apostrophe_as_single_quote = json_object.get("treatApostropheAsSingleQuote", False)
    language = json_object.get("language")

    if json_object.get("chineseTokenizer"):
        return ChineseBibleWordTokenizer()
    elif json_object.get("latinTokenizer"):
        return LatinWordTokenizer(treat_apostrophe_as_single_quote=treat_apostrophe_as_single_quote)
    elif json_object.get("latinWhiteSpaceIncludedTokenizer"):
        return LatinWhitespaceIncludedWordTokenizer(
            treat_apostrophe_as_single_quote=treat_apostrophe_as_single_quote,
            language=language,
            regex_rules_class=regex_rules_class
        )
    return None

def create_corpus(json_object, target_versification, psalm_superscription_tag):
    if "targetUsfmCorpusPath" in json_object:
        return UsfmFileTextCorpus(
                json_object["targetUsfmCorpusPath"],
            handler=ModifiedTextRowCollector,
            versification=target_versification,
            psalmSuperscriptionTag=psalm_superscription_tag
        )
    return UsxFileTextCorpus(
        json_object["targetUsxCorpusPath"],
        versification=target_versification
    )

def get_config(json_object):
    config_dict = {}

    regex_rules_class = DefaultRegexRules()
    if "regexRulesPath" in json_object:
        regex_rules_module = import_module_from_path("regex_rules", json_object["regexRulesPath"])
        regex_rules_class = getattr(regex_rules_module, "CustomRegexRules", DefaultRegexRules)()
    config_dict["regexRulesClass"] = regex_rules_class

    config_dict["tokenizer"] = create_tokenizer(json_object, regex_rules_class)

    target_versification_path = json_object.get("targetVersificationPath")
    config_dict["targetVersificationPath"] = target_versification_path
    config_dict["targetVersification"] = Versification.load(target_versification_path, fallback_name="web")

    config_dict["psalmSuperscriptionTag"] = json_object.get("psalmSuperscriptionTag", 'd')

    config_dict["projectName"] = json_object.get("projectName")
    if "tsvPath" in json_object:
        config_dict["tsvPath"] = json_object.get("tsvPath")
        config_dict["isCorpus"] = False
    else:
        config_dict["corpus"] = create_corpus(json_object, config_dict["targetVersification"], config_dict["psalmSuperscriptionTag"])
        config_dict["isCorpus"] = True
    
    config_dict["excludeBracketedText"] = json_object.get("excludeBracketedText", False)
    config_dict["excludeCrossReferences"] = json_object.get("excludeCrossReferences", False)
    config_dict["language"] = json_object.get("language")
    
    zw_removal_path = json_object.get("zwRemovalPath")
    config_dict["zwRemovalDf"] = pl.read_csv(zw_removal_path, separator='\t', infer_schema_length=0, quote_char=None) if zw_removal_path else None
    
    stop_words_path = json_object.get("stopWordsPath")
    config_dict["stopWordsDf"] = pl.read_csv(stop_words_path, separator='\t', infer_schema_length=0, quote_char=None) if stop_words_path else None
    
    return config_dict

def process_corpus(json_object):
    
    config_dict = get_config(json_object)

    if(config_dict["isCorpus"]):
        return{
            'corpus_to_tsv': corpus_to_tsv(
                targetVersification=config_dict["targetVersification"],
                sourceVersification=SOURCE_VERSIFICATION,
                corpus=config_dict["corpus"],
                tokenizer=config_dict["tokenizer"],
                project_name=config_dict["projectName"],
                excludeBracketedText=config_dict["excludeBracketedText"],
                excludeCrossReferences=config_dict["excludeCrossReferences"],
                language=config_dict["language"],
                zwRemovalDf=config_dict["zwRemovalDf"],
                stopWordsDf=config_dict["stopWordsDf"],
                regex_rules_class=config_dict["regexRulesClass"]
            )
        }
    else:
        return{
            'tokens_to_tsv': tokens_to_tsv(
                targetVersification=config_dict["targetVersification"],
                sourceVersification=SOURCE_VERSIFICATION,
                tsvPath=config_dict["tsvPath"],
                project_name=config_dict["projectName"],
                language=config_dict["language"],
                zwRemovalDf=config_dict["zwRemovalDf"],
                stopWordsDf=config_dict["stopWordsDf"],
                excludeBracketedText=config_dict["excludeBracketedText"],
                excludeCrossReferences=config_dict["excludeCrossReferences"],
                regex_rules_class=config_dict["regexRulesClass"]
            )            
        }
    
def get_json_data(json_file):
    with open(json_file) as file:
        json_data = json.load(file)

    solo_project = next((obj for obj in json_data if "solo" in obj), None)    
    if(solo_project is not None):
        json_data = [solo_project]
        
    return json_data

def main(json_path):

    json_data = get_json_data(json_path)

    with ProcessPoolExecutor() as executor:
        
        futures = {executor.submit(process_corpus, json_object): json_object for json_object in json_data}

        for future in as_completed(futures):
            try:
                result = future.result()
                print(f"Elapsed time: {time.time() - start:.2f} seconds")
            except Exception as e:
                print(f"Task generated an exception: {e}")

    print(f"Total elapsed time: {time.time() - start:.2f} seconds")

if __name__ == "__main__":
    main()


from machine.corpora import ParatextTextCorpus

from machine.tokenization import LatinWordTokenizer

from machine.translation import SymmetrizationHeuristic
from machine.translation import SymmetrizedWordAlignmentModelTrainer

from machine.translation.thot import ThotIbm1WordAlignmentModel
from machine.translation.thot import ThotWordAlignmentModelTrainer 
from machine.translation.thot import ThotWordAlignmentModelType
from machine.translation.thot import ThotSymmetrizedWordAlignmentModel


source_corpus = ParatextTextCorpus("data/VBL-PT")
target_corpus = ParatextTextCorpus("data/WEB-PT")
parallel_corpus = source_corpus.align_rows(target_corpus).tokenize(LatinWordTokenizer())

src_trg_trainer = ThotWordAlignmentModelTrainer(
    ThotWordAlignmentModelType.IBM1, parallel_corpus.lowercase(), "out/VBL-WEB-IBM1/src_trg"
)
trg_src_trainer = ThotWordAlignmentModelTrainer(
    ThotWordAlignmentModelType.IBM1, parallel_corpus.invert().lowercase(), "out/VBL-WEB-IBM1/trg_src"
)
symmetrized_trainer = SymmetrizedWordAlignmentModelTrainer(src_trg_trainer, trg_src_trainer)
symmetrized_trainer.train(lambda status: print(f"{status.message}: {status.percent_completed:.2%}"))
symmetrized_trainer.save()


src_trg_model = ThotIbm1WordAlignmentModel("out/VBL-WEB-IBM1/src_trg")
trg_src_model = ThotIbm1WordAlignmentModel("out/VBL-WEB-IBM1/trg_src")
symmetrized_model = ThotSymmetrizedWordAlignmentModel(src_trg_model, trg_src_model)
symmetrized_model.heuristic = SymmetrizationHeuristic.GROW_DIAG_FINAL_AND

segment_batch = list(parallel_corpus.lowercase().take(5))
alignments = symmetrized_model.align_batch(segment_batch)

for (source_segment, target_segment), alignment in zip(segment_batch, alignments):
    print("Source:", " ".join(source_segment))
    print("Target:", " ".join(target_segment))
    print("Alignment:", alignment)


import regex as re
from machine.scripture import (
    VerseRef,
    Versification,
)

def create_target_to_sources_dict(targetVersification:VerseRef):
    unused_versification_mapping = {}
    
    for key, value in targetVersification.mappings._standard_to_versification.items():
        if value.bbbcccvvvs in unused_versification_mapping:
            unused_versification_mapping[value.bbbcccvvvs].append(key)
        else:
            unused_versification_mapping[value.bbbcccvvvs] = [key]
    
    return unused_versification_mapping

def set_source_verse(targetVref:VerseRef, sourceVersification:Versification, unused_versification_mapping:dict[VerseRef:VerseRef]) -> tuple:
    
    sourceVref = ""
    source_verse_range_end = ""    
    
    mappings_to_targetVref = unused_versification_mapping.get(targetVref.bbbcccvvvs)
    
    if(mappings_to_targetVref != None):
        mappings_to_targetVref_len = len(mappings_to_targetVref)
               
        sourceVref = mappings_to_targetVref[0]
        if(mappings_to_targetVref_len > 1):
            last_mapping = mappings_to_targetVref[mappings_to_targetVref_len-1]
            source_verse_range_end = f"{re.sub(r'[^0-9]', '', last_mapping.bbbcccvvvs)}"[1:]
        
    if(sourceVref == ""):
        targetVref.change_versification(sourceVersification)
        sourceVref = targetVref
    
    return sourceVref, source_verse_range_end


from kathairo.tokenization.ChineseBibleWordTokenizer import ChineseBibleWordTokenizer
from machine.corpora import UsxFileTextCorpus
from machine.corpora import UsfmFileTextCorpus, UsxFileTextCorpus
from machine.tokenization import LatinWordTokenizer
from machine.scripture import (
    ENGLISH_VERSIFICATION,
    Versification
)

from machine.corpora import UsfmStylesheet

usfm = """\\id MAT - Test
\\h Matthew
\\mt Matthew
\\ip An introduction to Matthew
\\c 1
\\s Chapter One
\\p
\\v 1 This is verse \\pn one\\pn* of chapter one.
\\v 2 This is verse two\\f + \\fr 1:2: \\ft This is a footnote.\\f* of chapter one. 
""".replace("\n", "\r\n")

stylesheet = UsfmStylesheet("usfm.sty")

from machine.corpora import UsfmTokenizer, UsfmTokenType

usfm_tokenizer = UsfmTokenizer(stylesheet)
#tokens = usfm_tokenizer.tokenize(usfm)
#for token in tokens:
#  if token.type is UsfmTokenType.TEXT:
#    token.text = token.text.upper()

#print(usfm_tokenizer.detokenize(tokens))

from machine.corpora import UsfmParser

#usfm_parser = UsfmParser(stylesheet, usfm)
#state = usfm_parser.state
#while usfm_parser.process_token():
#  if state.token.type is UsfmTokenType.TEXT and state.is_verse_text:
#    state.token.text = state.token.text.upper()

#print(usfm_tokenizer.detokenize(state.tokens))

from machine.corpora import UsfmParserHandler

tokenizer = LatinWordTokenizer()
tokenized_row = tokenizer.tokenize("दाऊद ने मीकल से कहा, “ यहोवा, जिस ने तेरे पिता और उसके समस्त घराने के बदले मुझ को चुनकर अपनी प्रजा इस्राएल का प्रधान होने को ठहरा दिया है, उसके सम्मुख मैं ऐसा नाचा-- और मैं यहोवा के सम्मुख इसी प्रकार नाचा करूँगा ।")
for token in tokenized_row:
  print(token)
stop = True

class VerseTextUppercaser(UsfmParserHandler):
  def text(self, state, text):
    if state.is_verse_text:
      state.token.text = text.upper()
      print(state.token.text)

usfm_parser = UsfmParser(stylesheet=stylesheet, usfm=usfm, handler=VerseTextUppercaser())
usfm_parser.process_tokens()

print(usfm_tokenizer.detokenize(usfm_parser.state.tokens))

targetVersification = Versification.load("./resources/bsb_usx/release/versification.vrs", fallback_name="web")
corpus = UsfmFileTextCorpus("./resources/bsb_usfm", versification = targetVersification, include_markers=True)
tokenizer = LatinWordTokenizer()

row = corpus.get_rows

for row in corpus:
    tokenized_row = tokenizer.tokenize(row.text)
    if(row.ref.bbbcccvvvs[:6] == "019003"):
        print(row)

targetVersification = Versification.load("./resources/occb_simplified_usx/release/versification.vrs", fallback_name="web")
corpus = UsxFileTextCorpus("./resources/occb_simplified_usx/release/USX_1", versification = targetVersification)

for row in corpus:
    if(row.ref.bbbcccvvvs[:6] == "003006"):
        print(row)


import csv
from machine.utils import string_utils
from machine.corpora import UsxFileTextCorpus
from machine.corpora import ParatextTextCorpus, UsfmFileTextCorpus, UsxFileTextCorpus
from machine.tokenization import LatinWordTokenizer, WhitespaceTokenizer
from machine.scripture import (
    ENGLISH_VERSIFICATION,
    ORIGINAL_VERSIFICATION,
    RUSSIAN_ORTHODOX_VERSIFICATION,
    RUSSIAN_PROTESTANT_VERSIFICATION,
    SEPTUAGINT_VERSIFICATION,
    VULGATE_VERSIFICATION,
    LAST_BOOK,
    ValidStatus,
    VerseRef,
    Versification,
    get_bbbcccvvv,
)


import time
start_time=time.time()
#do something

from machine.corpora import ParatextTextCorpus
from machine.tokenization import LatinWordTokenizer

sourceVersification = Versification(name = "sourceVersification", base_versification=ENGLISH_VERSIFICATION)
targetVersification = Versification(name = "targetVersification", base_versification=ENGLISH_VERSIFICATION)

source_corpus = UsfmFileTextCorpus("./resources/bsb_usfm", versification = sourceVersification)
target_corpus = UsfmFileTextCorpus("./resources/arb-vd_usfm", versification = targetVersification)

parallel_corpus = source_corpus.align_rows(target_corpus).tokenize(LatinWordTokenizer())

from machine.translation import word_align_corpus
from machine.corpora import AlignedWordPair

aligned_corpus = word_align_corpus(parallel_corpus.lowercase())
#aligned_corpus = word_align_corpus(parallel_corpus.lowercase(), aligner="ibm1")

end_time=time.time()-start_time

for row in aligned_corpus.take(5):
    print("Source:", row.source_text)
    print("Target:", row.target_text)
    print("Alignment:", AlignedWordPair.to_string(row.aligned_word_pairs, include_scores=False))


from abc import abstractmethod
from typing import Generator, Optional

from machine.scripture.verse_ref import Versification
from machine.corpora.scripture_text import ScriptureText
from machine.corpora.stream_container import StreamContainer
from machine.corpora.text_row import TextRow
from kathairo.parsing.usx.usx_verse_parser import ModifiedUsxVerseParser


class UsxTextBase(ScriptureText):
    def __init__(self, id: str, versification: Optional[Versification]) -> None:
        super().__init__(id, versification)
        self._parser = ModifiedUsxVerseParser() #use modified parser

    @abstractmethod
    def _create_stream_container(self) -> StreamContainer:
        ...

    def _get_rows(self) -> Generator[TextRow, None, None]:
        with self._create_stream_container() as stream_container, stream_container.open_stream() as stream:
            for verse in self._parser.parse(stream):
                verse_ref = self._create_verse_ref(verse.chapter, verse.verse)
                yield from self._create_rows(verse_ref, verse.text, verse.is_sentence_start)



import csv
from machine.tokenization import WhitespaceTokenizer
from machine.corpora import ScriptureTextCorpus
from machine.scripture import VerseRef, Versification
import re
from kathairo.helpers.strings import is_unicode_punctuation, contains_number
from kathairo.helpers.paths import get_file_location
import os
from kathairo.helpers import strings as strings
from kathairo.helpers import versification
from pathlib import Path

def corpus_to_tsv(
    targetVersification:Versification, 
    sourceVersification:Versification, 
    corpus:ScriptureTextCorpus, 
    tokenizer:WhitespaceTokenizer, 
    project_name:str, 
    language:str, 
    zwRemovalDf:str, 
    stopWordsDf:str, 
    excludeBracketedText:bool = False, 
    excludeCrossReferences:bool = False, 
    regex_rules_class = None
):    

    corpus_array = corpus_to_array(targetVersification, sourceVersification, corpus, tokenizer)

    #TODO try to make these parallel sometime
    array_to_token_level_tsv(corpus_array, project_name, language, zwRemovalDf, stopWordsDf, excludeBracketedText, excludeCrossReferences, regex_rules_class)
    array_to_verse_level_tsv(corpus_array, project_name, language)
  
def corpus_to_array(targetVersification:Versification, sourceVersification:Versification, corpus:ScriptureTextCorpus, tokenizer:WhitespaceTokenizer):
    
    corpus_array = []
    unused_versification_mapping = versification.create_target_to_sources_dict(targetVersification)
    
    verse_range_list = []
    append_range_list = verse_range_list.append

    pattern = re.compile(r'[^0-9]')

    for tokenized_row, untokenized_row in zip(corpus.tokenize(tokenizer), corpus):
        row_tokens = []
        
        for index in range(len(tokenized_row.segment)):
        
            token = tokenized_row.segment[index]
            row_tokens.append(token)
        
        targetVref = VerseRef.from_bbbcccvvv(untokenized_row.ref.bbbcccvvv, targetVersification)
        sourceVref, source_verse_range_end = versification.set_source_verse(targetVref, sourceVersification, unused_versification_mapping)

        if not untokenized_row.is_in_range or untokenized_row.is_range_start:
            for verse_range_row in verse_range_list:
                verse_range_row[3] = rowBcv
                verse_range_row[4] = sourceBcv
                corpus_array.append(verse_range_row)
            verse_range_list.clear()

        sourceBcv = f"{re.sub(r'[^0-9]', '', sourceVref.bbbcccvvvs)}"[1:]
        rowBcv = f"{re.sub(r'[^0-9]', '', untokenized_row.ref.bbbcccvvvs)}"[1:]
        
        row_text = untokenized_row.text
        if row_text:
            if untokenized_row.is_in_range:
                append_range_list([rowBcv, sourceBcv, row_tokens, "", source_verse_range_end, row_text])
            else:
                corpus_array.append([rowBcv, sourceBcv, row_tokens, "", source_verse_range_end, row_text])
        
    return corpus_array

def tokens_to_tsv(
    targetVersification:Versification, 
    sourceVersification:Versification, 
    tsvPath:str, 
    project_name:str, 
    language:str, 
    zwRemovalDf:str, 
    stopWordsDf:str, 
    excludeBracketedText:bool = False, 
    excludeCrossReferences:bool = False, 
    regex_rules_class = None
):
    corpus_array = tokens_to_array(tsvPath, targetVersification, sourceVersification)
    
    #TODO try to make these parallel sometime
    array_to_verse_level_tsv(corpus_array, project_name, language)
    array_to_token_level_tsv(corpus_array, project_name, language, zwRemovalDf, stopWordsDf, excludeBracketedText, excludeCrossReferences, regex_rules_class)

def tokens_to_array(tsvPath:str, targetVersification:Versification, sourceVersification:Versification,):
    pattern = re.compile(r'[^0-9]')
    unused_versification_mapping = versification.create_target_to_sources_dict(targetVersification)

    input_file = Path(tsvPath)
    corpus = [r for r in csv.DictReader(input_file.open("r", encoding='utf-8'), delimiter="\t", quoting=csv.QUOTE_NONE, quotechar=None)]
    
    corpus_array = []
    corpus_row = []
    verse_tokens = []
    verse_text = ""
    bcv_id = "40001001"
    previous_bcv_id = "40001001"
    for row in corpus:
        
        previous_bcv_id = bcv_id
        bcv_id = row['id'][0:8]
        if(bcv_id != previous_bcv_id):
            targetVref = VerseRef.from_bbbcccvvv(int(previous_bcv_id))
            sourceVref, source_verse_range_end = versification.set_source_verse(targetVref, sourceVersification, unused_versification_mapping)
            sourceBcv = f"{re.sub(r'[^0-9]', '', sourceVref.bbbcccvvvs)}"[1:]
            
            corpus_row = [previous_bcv_id, sourceBcv, verse_tokens, "", "", verse_text]
            corpus_array.append(corpus_row)
            corpus_row = []
            verse_tokens = []
            verse_text=""
        
        token = row['text']
        #token = [row['id'], row['text'], row['isPunc'], row['isPrimary']]    
        
        verse_tokens.append(token)
        
        verse_text += token
        #verse_text += row['text']
    
    #for final row
    targetVref = VerseRef.from_bbbcccvvv(int(previous_bcv_id))
    sourceVref, source_verse_range_end = versification.set_source_verse(targetVref, sourceVersification, unused_versification_mapping)
    sourceBcv = f"{re.sub(r'[^0-9]', '', sourceVref.bbbcccvvvs)}"[1:]

    corpus_row = [previous_bcv_id, sourceBcv, verse_tokens, "", "", verse_text]
    corpus_array.append(corpus_row)
    
    return corpus_array

def array_to_verse_level_tsv(
    corpus_array, 
    project_name: str, 
    language: str
):
    outputFileName = get_file_location("output", language, project_name, "verse", "verse")
    os.makedirs(os.path.dirname(outputFileName), exist_ok=True)
    with open(outputFileName, 'w', newline='', encoding='utf-8') as out_file:
        
        tsv_writer = csv.writer(out_file, delimiter='\t', quoting=csv.QUOTE_NONE, quotechar=None)
        tsv_writer.writerow(["id", "source_verse", "text", "id_range_end", "source_verse_range_end"])
        
        for row in corpus_array:
            tsv_writer.writerow([row[0], row[1], row[5], row[3], row[4]])
            
def array_to_token_level_tsv(corpus_array,
                project_name:str, language:str, zwRemovalDf:str, stopWordsDf:str, excludeBracketedText:bool = False, 
                excludeCrossReferences:bool = False, regex_rules_class = None):
    
    zw_removal_df= zwRemovalDf
        
    stop_words_df=stopWordsDf

    WORD_LEVEL_PUNCT_REGEX = regex_rules_class.WORD_LEVEL_PUNCT_REGEX
    
    outputFileName = get_file_location("output", language, project_name, "token", "token")
    os.makedirs(os.path.dirname(outputFileName), exist_ok=True)
    with open(outputFileName, 'w', newline='', encoding='utf-8') as out_file:
        tsv_writer = csv.writer(out_file, delimiter='\t', quoting=csv.QUOTE_NONE, quotechar=None)

        tsv_writer.writerow(["id", "source_verse", "text", "skip_space_after", "exclude", "id_range_end", "source_verse_range_end", "required"])

        in_brackets = False
        
        in_parentheses = False
        is_cross_reference = False
        has_number = False
        unprinted_parenthetical_tokens = []
        
        previous_verse_num = 0
        previous_skip_space_after = ""
        
        for row in corpus_array:

            current_verse_num = row[0]
            if(current_verse_num != previous_verse_num):
                wordIndex = 1
            previous_verse_num = current_verse_num

            for index in range(len(row[2])):
            
                has_number, is_cross_reference, unprinted_parenthetical_tokens = process_and_print_parenthetical_tokens(
                    excludeCrossReferences, 
                    tsv_writer, 
                    in_parentheses, 
                    has_number,
                    is_cross_reference, 
                    unprinted_parenthetical_tokens
                )

                #id = row[2][index][0]
                
                token = row[2][index]
                #token = row[2][index][1]
                
                #isPunc = row[2][index][2]
                #if(isPunc == "False"):
                #    isPunc = ""
                #else:
                #    isPunc = "y"
                #isPrimary = row[2][index][3]
                #if(isPrimary == "False"):
                #    isPrimary = "n"
                #else:
                #    isPrimary = "y"
                         
                    
                #Getting Tokens
                previous_previous_token, previous_token, next_token, next_next_token = get_surrounding_tokens(row[2], index)
                
                #Skip Space After
                skip_space_after = "y"
                
                if(is_stop_word(stop_words_df, token)):
                    continue 
                elif((next_token in strings.spaces) or (next_token in strings.invisible_spaces and next_next_token in strings.spaces)):
                    skip_space_after = ""
                
                #''' 
                #ZW Characters
                token = remove_zw_characters(zw_removal_df, token)

                #Exclude
                exclude = ""

                if isinstance(token, str) and len(token) > 0 and all(in_brackets or is_unicode_punctuation(char) for char in token):
                    exclude = "y"

                    if is_token_word_level_punct(stop_words_df, WORD_LEVEL_PUNCT_REGEX, previous_skip_space_after, token, previous_previous_token, previous_token, next_token, next_next_token, skip_space_after):
                        exclude = ""
                    
                if(excludeBracketedText and '[' in token):
                    in_brackets = True
                    exclude = "y"
                if(']' in token):
                    in_brackets = False
                    
                if(excludeCrossReferences and '(' in token): #add to unit test to look for that all things marked as cross references are indeed cross-references and no token has a colon and a parentheses
                    in_parentheses = True
                if(excludeCrossReferences and in_parentheses and contains_number(token)):#add this change to the unit test
                    has_number = True
                if(excludeCrossReferences and in_parentheses and has_number and ':' in token):
                    is_cross_reference = True
                
                #Required
                required = calculate_required(token)
                #'''
                
                #Printing
                wordIndexStr = str(wordIndex).zfill(3)
                
                if(row[5] != ""):
                    if(in_parentheses):
                        unprinted_parenthetical_tokens.append(([f"{row[0]}{wordIndexStr}", f"{row[1]}", token, skip_space_after, exclude,  row[3], row[4], required]))
                        #unprinted_parenthetical_tokens.append(([id, f"{row[1]}", token, skip_space_after, isPunc,  row[3], row[4], isPrimary]))
                    else:
                        tsv_writer.writerow([f"{row[0]}{wordIndexStr}", f"{row[1]}", token, skip_space_after, exclude, row[3], row[4], required])
                        #tsv_writer.writerow([id, f"{row[1]}", token, skip_space_after, isPunc, row[3], row[4], isPrimary])
                
                if(')' in token):
                    in_parentheses = False
                
                wordIndex += 1 
                previous_skip_space_after = skip_space_after

def process_and_print_parenthetical_tokens(excludeCrossReferences, tsv_writer, in_parentheses, has_number, is_cross_reference, unprinted_parenthetical_tokens):
    if(not in_parentheses):    
        for unprinted_cross_reference_token in unprinted_parenthetical_tokens:
            if(excludeCrossReferences and is_cross_reference):
                unprinted_cross_reference_token[4] = 'y' #exclude if is_cross_reference
            tsv_writer.writerow(unprinted_cross_reference_token)
        has_number = False
        is_cross_reference = False
        unprinted_parenthetical_tokens.clear()
    return has_number, is_cross_reference, unprinted_parenthetical_tokens

def get_surrounding_tokens(token_array, index):
    segment = token_array
    max_segment_index = len(segment) - 1

    previous_previous_token = segment[index - 2] if index - 2 >= 0 else ''
    previous_token = segment[index - 1] if index - 1 >= 0 else ''
    next_token = segment[index + 1] if index + 1 <= max_segment_index else strings.space
    next_next_token = segment[index + 2] if index + 2 <= max_segment_index else strings.space

    return previous_previous_token, previous_token, next_token, next_next_token

def is_stop_word(stop_words_df, token):
    return token in strings.stop_words or (stop_words_df is not None and token in stop_words_df["stop_words"].to_numpy())      

def remove_zw_characters(zw_removal_df, token):
    if(zw_removal_df is not None and token in zw_removal_df["words"].to_numpy()):
        token = token.replace(strings.zwsp, strings.empty_string).replace(strings.zwj, strings.empty_string).replace(strings.zwnj, strings.empty_string)
    return token

def is_token_word_level_punct(stop_words_df, WORD_LEVEL_PUNCT_REGEX, previous_skip_space_after, token, previous_previous_token, previous_token, next_token, next_next_token, skip_space_after):
    next_substring_token = (
        strings.space + next_token if skip_space_after == ""
        else next_next_token if is_stop_word(stop_words_df, next_token)
        else next_token
    )

    previous_substring_token = (
        previous_token + strings.space if previous_skip_space_after == ""
        else previous_previous_token if is_stop_word(stop_words_df, previous_token)
        else previous_token
    )

    substring = previous_substring_token + token + next_substring_token
    index = len(previous_substring_token + token) - 1

    return WORD_LEVEL_PUNCT_REGEX.match(substring, index) is not None

def calculate_required(token):
    return "y" if any(not is_unicode_punctuation(char) for char in token) else "n"         


from usfm_grammar import USFMParser
from lxml import etree

def print_errors_for_usx(file_path:str):
    with open(file_path, 'r', encoding='utf-8') as usx_file:
        usx_str = usx_file.read()
        usx_obj = etree.fromstring(bytes(usx_str, encoding='utf8'))
        my_parser = USFMParser(from_usx=usx_obj)
        return my_parser

def print_errors_for_usfm(file_path:str):    
    input_usfm_str = open(file_path,"r", encoding='utf8').read()
    my_parser = USFMParser(input_usfm_str)
    return my_parser

import os

#directory_path = "./resources/bsb_usx/release/USX_1/"
directory_path = "./resources/ru_rsb/"
file_list = os.listdir(directory_path)

for file in file_list:
    #my_parser = print_errors_for_usx(directory_path+file)
    my_parser = print_errors_for_usfm(directory_path+file)
    
    errors = my_parser.errors
    if(errors is not None):
        for error in errors:
            #print(error)

            with open('output_russian.txt', 'a') as f:
                f.write(str(error))   
                f.write("\n")    


#--------------------------------------------

#test_xml_file = "./resources/bsb_usx/release/USX_1/1CH.usx"
#test_usfm_file = "./resources/bsb_usfm/50EPHBSB.SFM"

#my_parser = print_errors_for_usx(test_xml_file)
#my_parser = print_errors_for_usfm(test_usfm_file)

#errors = my_parser.errors
#for error in errors:
#    print(error)

#--------------------------------------------

# print(my_parser.to_usj())
# print(my_parser.to_list())



import polars as pl
import csv

book_abbrev_dict = {
    1:'GEN',
    2:'EXO',
    3:'LEV',
    4:'NUM',
    5:'DEU',
    6:'JOS',
    7:'JDG',
    8:'RUT',
    9:'1SA',
    10:'2SA',
    11:'1KI',
    12:'2KI',
    13:'1CH',
    14:'2CH',
    15:'EZR',
    16:'NEH',
    17:'EST',
    18:'JOB',
    19:'PSA',
    20:'PRO',
    21:'ECC',
    22:'SNG',
    23:'ISA',
    24:'JER',
    25:'LAM',
    26:'EZK',
    27:'DAN',
    28:'HOS',
    29:'JOL',
    30:'AMO',
    31:'OBA',
    32:'JON',
    33:'MIC',
    34:'NAM',
    35:'HAB',
    36:'ZEP',
    37:'HAG',
    38:'ZEC',
    39:'MAL',
    40:'MAT',
    41:'MRK',
    42:'LUK',
    43:'JHN',
    44:'ACT',
    45:'ROM',
    46:'1CO',
    47:'2CO',
    48:'GAL',
    49:'EPH',
    50:'PHP',
    51:'COL',
    52:'1TH',
    53:'2TH',
    54:'1TI',
    55:'2TI',
    56:'TIT',
    57:'PHM',
    58:'HEB',
    59:'JAS',
    60:'1PE',
    61:'2PE',
    62:'1JN',
    63:'2JN',
    64:'3JN',
    65:'JUD',
    66:'REV',
}

df = pl.read_csv("nt_CUVMPS_verse.tsv", separator='\t', infer_schema_length=0)

previous_chapter_id = 0
previous_book_id = 0

book_arr = []

for row in df.iter_rows(named=True):
    id = row["id"]
    verse = str(row["text"])

    book_id = id[0:2]
    chapter_id = id[2:5]
    verse_id = id[5:8]
    

    
    if(previous_book_id != book_id):
        if(book_arr != []):
            with open(output_file_name, 'w', newline='', encoding='utf-8') as out_file:
                tsv_writer = csv.writer(out_file, delimiter='\t') # quoting=csv.QUOTE_NONE, quotechar=None
                tsv_writer.writerows(book_arr)
            
        book_arr = []
        #df[row_index, "text"] = f"\\id {book_abbrev} \\c {int(chapter_id)} \\v {int(verse_id)} {verse}"
        book_abbrev = book_abbrev_dict[int(book_id)]
        output_file_name = f"SFM/{book_id}_{book_abbrev}.SFM"
        
        book_arr.append([f"\\id {book_abbrev} \\c {int(chapter_id)} \\v {int(verse_id)} {verse}"])
    elif(previous_chapter_id != chapter_id):
        #df[row_index, "text"] = f"\\c {int(chapter_id)} \\v {int(verse_id)} {verse}"
        book_arr.append([f"\\c {int(chapter_id)} \\v {int(verse_id)} {verse}"])
    else:
        #df[row_index, "text"] = f"\\v {int(verse_id)} {verse}"
        book_arr.append([f"\\v {int(verse_id)} {verse}"])
    
    previous_chapter_id = chapter_id
    previous_book_id = book_id

with open(output_file_name, 'w', newline='', encoding='utf-8') as out_file:
    tsv_writer = csv.writer(out_file, delimiter='\t') # quoting=csv.QUOTE_NONE, quotechar=None
    tsv_writer.writerows(book_arr)

#df.select(["text"]).write_csv(output_file_name, separator='\t')